'use strict';


/**
 * Modules: Node
 */
const childProcess = require('child_process'),
    path = require('path');


/**
 * Modules: external
 */
const glob = require('glob'),
    _ = require('lodash'),    
    publishRelease = require('publish-release'),
    progressBar = require('progress');


/**
 * Modules: internal
 */
const appRoot = path.resolve('.'),
    packageJson = require(appRoot + '/package.json'),
    logger = require(appRoot + '/lib/logger');


/**
 * Create Asset list
 * @returns {Array} List of absolute paths to files to be published
 */
let createAssetList = function() {
    return glob.sync(path.join(appRoot, packageJson.build.directoryRelease, '*.zip'));
};


/**
 * Github release options
 * @param {Array} assetList - Files to publish
 * @param {String} githubToken - Valid Github token
 * @returns {Object|undefined} - 'publish-release' module settings object
 */
let createGithubOptions = function(assetList, githubToken) {
    if (!assetList || !_.isArray(assetList)) {
        logger.logErr('error', 'assetList missing or wrong format.');
        return;
    }

    if (_.isEmpty(assetList)) {
        logger.logErr('error', 'assetList is empty.');
        return;
    }

    if (!githubToken || !_.isString(githubToken)) {
        logger.logErr('error', 'githubToken missing or wrong format.');
        return;
    }

    return {
        assets: assetList,
        token: githubToken,
        owner: packageJson.author.name,
        repo: packageJson.name,
        tag: 'v' + packageJson.version,
        name: packageJson.build.productName + ' ' + 'v' + packageJson.version,
        notes: packageJson.build.productName + ' ' + 'v' + packageJson.version,
        draft: true,
        reuseRelease: true,
        reuseDraftOnly: false,
        prerelease: false,
        target_commitish: 'master'
    };
};


/**
 * Handle events emitted by PublishRelease
 * @param {PublishRelease} release - PublishRelease object
 * @returns {Boolean|undefined} - Result of add event handlers
 */
let addEventHandlers = function(release) {
    if (!release || !_.isObject(release)) {
        logger.logErr('error', 'release missing or wrong format.');
        return;
    }

    let bar;

    release.on('created-release', function() {
        logger.log('Github Release', 'created');
    });

    release.on('reuse-release', function() {
        logger.log('Github Release', 'exists');
    });

    release.on('upload-asset', function(name) {
        logger.log('Github Upload', 'starting');
    });

    release.on('uploaded-asset', function(name) {
        if (bar) { bar.update(1); }  
        logger.log('Github Upload', 'complete');
    });

    release.on('upload-progress', function(name, event) {
        if (!bar) { bar = new progressBar(name + ' [:bar] :percent (ETA: :etas)', { complete: '=', incomplete: ' ', width: 30, total: event.length }); }
        if (bar) { bar.tick(event.delta); }         
    });

    return true;
};


/** Returns with a callback holding the global settings object for passed OSX app bundle ids.
 * @param {Function} cb - Callback
 */
let releaseOnGithub = function(cb) {
    // Init callback
    let callback = cb || function() {};

    // Create Config
    let options = createGithubOptions(createAssetList(), process.env.GITHUB_TOKEN);

    // Type check
    if (!options || !_.isObject(options)) {
        return callback(new Error('options missing or wrong format.'));
    }

    // Create Release
    let release = publishRelease(options, function(err, result) {
        if (err) {
            return callback(err);
        }
        return callback(null, result);
    });

    addEventHandlers(release);
};


/**
 * Initialize main process if called from CLI
 */
if (require.main === module) {
    releaseOnGithub(function(err, result) {
        if (err) {
            logger.logErr('error', err);
            return process.exit(1);
        }
        logger.log('Github Release', 'complete:', result['url']);
        return process.exit(0);
    });    
}


/**
 * exports
 */
module.exports = {
    assetList: createAssetList,
    createGithubOptions: createGithubOptions,
    release: releaseOnGithub
};
